import matplotlib.pyplot as plt
import numpy as np
import time
import sys
import random


###### PARTIE 1 ######

def construction_mphf(set_kmer, n, gamma=2, nb_niveaux=3):
    """
    Construit une fonction de hachage minimale parfaite (MPHF) pour un ensemble de k-mers.

    Parameters:
    set_kmer (set): Ensemble de k-mers.
    n (int): Taille de l'ensemble de k-mers.
    gamma (int): Facteur de réduction de la taille de la table de hachage. Default: 2.
    nb_niveaux (int): Nombre de niveaux de réduction de la taille de la table de hachage. Default: 3.

    Returns:
    mphf (list): Table de hachage minimale parfaite.

    Examples:
    >>> set_kmer = {str(i) for i in range(10)}
    >>> n = 10
    >>> mphf = construction_mphf(set_kmer, n)
    >>> len(mphf) == n
    True
    >>> all(0 <= mphf[i] < n for i in range(n))
    True
    >>> len(mphf) == n
    True
    """
    # Initialisation

    set_kmer_courant = set_kmer.copy()
    tableaux = []
    collision = set()
    for _ in range(nb_niveaux):
        if len(set_kmer_courant) > 0:
            l = len(set_kmer_courant)
            tableau_principal = [-1] * (gamma * l)
            for kmer in set_kmer_courant:
                hachage = abs(hash(kmer))# positif
                adresse = hachage % len(tableau_principal)
                if tableau_principal[adresse] != -1:  # Collision
                    collision.add(kmer)
                else:
                    tableau_principal[adresse] = hachage  # Stocker le hachage
            tableaux.append(tableau_principal) # on conserve les positions sans collision dans notre table finale

            set_kmer_courant = collision.copy()  # on prépare les positions en collision dans le set
            # courant pour un nouveau hachage
            collision = set()  # on replace les positions non-hachées en tant que set pour le hachage

    # Construction de la MPHF
    grand_tableau = []
    for tableau in tableaux:
        grand_tableau.extend(tableau)  # on ajoute tous les éléments de tous les tabeaux dans le tableau final, sans renvoyer de valeur
    count = 0
    list_h_pos = []
    for pos in range(len(grand_tableau)):
        h_tab = grand_tableau[pos]
        if h_tab != -1:
            count += 1
            list_h_pos.append((h_tab, count)) # fonction de hashage minimale parfaite !

    mphf = list_h_pos

    '''col_h = []
    col_kmer = []
    for k in kmer:
        h = abs(hash(k))
        ok = 0
        for e in range(len(list_h_pos)):
            if list_h_pos[e][0] == h:
                ok += 1
        if ok == 0:
            col_h.append(h)  # si besoin d'une liste de h qui sont pas dans tableau et qu'il faut rajouter apres
            col_kmer.append(k)  # pareil mais avec les kmer, jsp ce qu'on aura besoin ensuite
        # compléter
    # hacher le kmer
    # si le hash est dans le grand_tableau
    # récupérer son index
    # récupérer son rang (utiliser la fonction count())
    # ajouter à la mphf [h, rang]
    # mettre à jour max_rang

    for kmer in set_kmer_courant:  # permet de récupérer la dernière collision si elle existe
        max_rang += 1
        h = abs(hash(kmer))
        mphf.append([h, max_rang])
'''

    return mphf


def get_hash_mphf(mphf, kmer):
    """
    Calcule le hash d'un k-mer à l'aide d'une fonction de hachage minimale parfaite (MPHF).

    Parameters:
    mphf (list): Table de hachage minimale parfaite.
    kmer (str): K-mer à hasher.

    Returns:
    int: Hash du k-mer.

    Examples:
    >>> set_kmer = {str(i) for i in range(10)}
    >>> n = 10
    >>> mphf = construction_mphf(set_kmer, n)
    >>> kmer = "5"
    >>> hash_mphf = get_hash_mphf(mphf, kmer)
    >>> 0 <= hash_mphf < n
    True
    """
    hashkmer = abs(hash(kmer))
    for hach in mphf:
        if hach[0] == hashkmer:
            return hashkmer
        else:
            return "not in the table"


def create_hash_table(set_kmer, n):
    """
    Crée une table de hachage à partir d'un ensemble de k-mers et d'une mphf

    Parameters:
    set_kmer (set): Ensemble de k-mers.
    n (int): Taille de la table de hachage.

    Returns:
    list: Table de hachage créée à partir des k-mers
    mphf: la mphf

    Examples:
    >>> set_kmer = {'ATCG', 'TGCA', 'GCTA'}
    >>> n = 10
    >>> tableau = create_hash_table(set_kmer, n)
    >>> len(tableau) == n
    True
    >>> all(kmer in tableau for kmer in set_kmer)
    True
    """
    mphf = construction_mphf(set_kmer, n)
    liste = [0] * n
    n = 0
    for kmer in mphf:
        liste[n] = kmer[0]
        n += 1
    return liste, mphf
    # créer la mphf pour les kmers
    # initialiser un tableau de taille n (une liste)
    # écrire les kmers aux adresses du tableau données par la mphf
    # retourner le tableau et la mphf


def generer_kmers(n, k):
    '''
    genere un set de n k-mers
    '''
    kmers = set()
    while len(kmers) < n:
        kmer = ''.join(random.choice('ATCG') for _ in range(k))
        kmers.add(kmer)
    return kmers


def compare_taille(n_max, fichier_sortie):
    """
    Crée un graphique de la taille de la table de hachage comparée à la taille d'un dictionnaire pour stocker jusqu'à n-max kmers
    :param n_max: nombre maximum de kmers traités
    :param fichier_sortie: chemin dans l'arborescence de fichiers pour récupérer le graphique
    :return: outfile
    """
    n_values = []  # table pour les plots
    table_size = []  # table pour la mphf
    dict_size = []  # table pour le dictionnaire
    k = 21  # taille des kmers générés

    for n in range(100, n_max, 1000):
        set_kmer = generer_kmers(n, k)  # génération de jusqu'à n_max kmers
        tableau, mphf = create_hash_table(set_kmer, n)  # création de la table de hachage

        n_values.append(n)  # valeurs de l'axe des abscisse
        table_size.append(sys.getsizeof(tableau) + sys.getsizeof(
            mphf))  # pourquoi ici on ne mesure pas juste la taille en mémoire du tableau ? Car on a également besoin
        # de garder en mémoire la mphf, à cause des collisions 
        dict_size.append(sys.getsizeof(set_kmer))

    # création du plot
    plt.plot(n_values, table_size, label='Table avec MPHF')
    plt.plot(n_values, dict_size, label='Dict')
    plt.xlabel('n')
    plt.xticks(n_values, [str(x) for x in n_values], rotation=45)
    plt.ylabel('Taille (octets)')
    plt.title('Évolution de la taille de la table de hachage avec MPHF et du dict')
    plt.legend()
    plt.savefig(fichier_sortie)
    plt.close()

if __name__ == "__main__":
    import doctest
    doctest.testmod()
    kmers = generer_kmers(10, 10)
    set_kmer = set(kmers)
    n = 10
    mphf = construction_mphf(set_kmer, n)
    print('mpfh')
    print(mphf)
    print('get')
    print(get_hash_mphf(mphf, "chenille"))
    print("hash table")
    n = len(mphf)
    print(create_hash_table(set_kmer, n))
    print("compare_taille")
    print(compare_taille(10000, "/home/m1miso/pacome.riobe.etu/PycharmProjects/pythonProjectsdd2/sortie.png"))